from copy import deepcopy
import logging

# 3rd-party
from lxml import etree
from lxml.builder import E

# local
from jnpr.junos.factory.table import Table
from jnpr.junos.jxml import remove_namespaces, remove_namespaces_and_spaces
from jnpr.junos.decorators import checkSAXParserDecorator

logger = logging.getLogger("jnpr.junos.factory.optable")


class OpTable(Table):

    # -------------------------------------------------------------------------
    # PUBLIC METHODS
    # -------------------------------------------------------------------------

    @checkSAXParserDecorator
    def get(self, *vargs, **kvargs):
        """
        Retrieve the XML table data from the Device instance and
        returns back the Table instance - for call-chaining purposes.

        If the Table was created with a :path: rather than a Device,
        then this method will load the XML from that file.  In this
        case, the \*vargs, and \**kvargs are not used.

        ALIAS: __call__

        :vargs:
          [0] is the table :arg_key: value.  This is used so that
          the caller can retrieve just one item from the table without
          having to know the Junos RPC argument.

        :kvargs:
          these are the name/value pairs relating to the specific Junos
          XML command attached to the table.  For example, if the RPC
          is 'get-route-information', there are parameters such as
          'table' and 'destination'.  Any valid RPC argument can be
          passed to :kvargs: to further filter the results of the :get():
          operation.  neato!

        NOTES:
          If you need to create a 'stub' for unit-testing
          purposes, you want to create a subclass of your table and
          overload this methods.
        """
        self._clearkeys()

        if self._path is not None:
            # for loading from local file-path
            self.xml = remove_namespaces(etree.parse(self._path).getroot())
            return self

        if self._lxml is not None:
            return self

        argkey = vargs[0] if len(vargs) else None

        rpc_args = {}

        if self._use_filter:
            try:
                filter_xml = generate_sax_parser_input(self)
                rpc_args['filter_xml'] = filter_xml
            except Exception as ex:
                logger.debug("Not able to create SAX parser input due to "
                               "'%s'" % ex)

        self.D.transform = lambda: remove_namespaces_and_spaces
        rpc_args.update(self.GET_ARGS)    # copy default args
        # saltstack get_table pass args as named keyword
        if 'args' in kvargs and isinstance(kvargs['args'], dict):
            rpc_args.update(kvargs.pop('args'))
        rpc_args.update(kvargs)           # copy caller provided args


        if hasattr(self, 'GET_KEY') and argkey is not None:
            rpc_args.update({self.GET_KEY: argkey})

        # execute the Junos RPC to retrieve the table
        self.xml = getattr(self.RPC, self.GET_RPC)(**rpc_args)

        # returning self for call-chaining purposes, yo!
        return self


def generate_sax_parser_input(obj):
    """
    Used to generate xml object from Table/view to be used in SAX parsing
    Args:
        obj: self object which contains table/view details

    Returns: lxml etree object to be used as sax parser input

    """
    if '/' in obj.ITEM_XPATH:
        tags = obj.ITEM_XPATH.split('/')
        parser_ingest = E(tags.pop(-1), E(obj.ITEM_NAME_XPATH))
        for tag in tags[::-1]:
            parser_ingest = E(tag, parser_ingest)
    else:
        parser_ingest = E(obj.ITEM_XPATH, E(obj.ITEM_NAME_XPATH))
    local_field_dict = deepcopy(obj.VIEW.FIELDS)
    # first make element out of group fields
    if obj.VIEW.GROUPS:
        for group, group_xpath in obj.VIEW.GROUPS.items():
            # need to pop out group items so that it wont be reused with fields
            group_field_dict = ({k: local_field_dict.pop(k)
                                 for k, v in obj.VIEW.FIELDS.items()
                                 if v.get('group') == group})
            group_ele = E(group_xpath)
            for key, val in group_field_dict.items():
                group_ele.append(E(val.get('xpath')))
            parser_ingest.append(group_ele)
    for i, item in enumerate(local_field_dict.items()):
        # i is the index and item will be taple of field key and value
        field_dict = item[1]
        if 'table' in field_dict:
            # handle nested table/view
            child_table = field_dict.get('table')
            parser_ingest.insert(i + 1, generate_sax_parser_input(child_table))
        else:
            xpath = field_dict.get('xpath')
            # xpath can be multi level, for ex traffic-statistics/input-pps
            if '/' in xpath:
                tags = xpath.split('/')
                obj = E(tags[0])
                for tag in tags[1:]:
                    obj.append(E(tag))
                parser_ingest.insert(i + 1, obj)
            else:
                parser_ingest.insert(i + 1, E(xpath))
    return parser_ingest



from jnpr.junos.exception import ConnectNotMasterError
from jnpr.junos.exception import RpcError


def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {'RE_hw_mi': "(Routing Engine hardware multi-instance) A boolean "
                        "indicating if this is a multi-chassis system.",
            'serialnumber':  "A string containing the serial number of the "
                             "device's chassis. If there is no chassis serial "
                             "number, the serial number of the backplane or "
                             "midplane is returned.", }


def get_facts(device):
    """
    Gathers facts from the <get-chassis-inventory/> RPC.
    """
    rsp = device.rpc.get_chassis_inventory(normalize=True)
    if rsp.tag == 'error':
        raise RpcError()

    if (rsp.tag == 'output' and
       rsp.text.find('can only be used on the master routing engine') != -1):
        # An error; due to the fact that this RPC can only be executed on the
        # master Routing Engine
        raise ConnectNotMasterError()

    RE_hw_mi = False
    if rsp.tag == 'multi-routing-engine-results':
        RE_hw_mi = True

    serialnumber = (
        rsp.findtext('.//chassis[1]/serial-number') or
        rsp.findtext('.//chassis-module[name="Backplane"]/serial-number') or
        rsp.findtext('.//chassis-module[name="Midplane"]/serial-number'))

    return {'RE_hw_mi': RE_hw_mi,
            'serialnumber': serialnumber, }



import sys
import os
import yaml
import types

from jnpr.junos.factory.factory_loader import FactoryLoader

__all__ = []


class MetaPathFinder(object):
    def find_module(self, fullname, path=None):
        mod = fullname.split('.')[-1]
        if mod in [os.path.splitext(i)[0] for i in os.listdir(
                os.path.dirname(__file__))]:
            return MetaPathLoader()


class MetaPathLoader(object):
    def load_module(self, fullname):
        if fullname in sys.modules:
            return sys.modules[fullname]
        mod = fullname.split('.')[-1]
        modObj = types.ModuleType(mod,
                                  'Module created to provide a context for %s'
                                  % mod)
        with open(os.path.join(os.path.dirname(__file__), mod + '.yml'), 'r')\
                as stream:
            try:
                modules = FactoryLoader().load(yaml.load(stream,
                                                         Loader=yaml.FullLoader))
            except yaml.YAMLError as exc:
                raise ImportError("%s is not loaded" % mod)
        for k, v in modules.items():
            setattr(modObj, k, v)
        sys.modules[fullname] = modObj
        return modObj

sys.meta_path.insert(0, MetaPathFinder())



from jnpr.junos.utils.fs import FS
from jnpr.junos.exception import RpcError
from jnpr.junos.jxml import INHERIT
from lxml.builder import E


def facts_domain(junos, facts):
    """
    The following facts are required:
        facts['hostname']

    The following facts are assigned:
        facts['domain']
        facts['fqdn']
    """

    try:
        domain_filter_xml = E('configuration', E('system', E('domain-name')))
        domain = junos.rpc.get_config(filter_xml=domain_filter_xml,
                                      options=INHERIT)
        domain_name = domain.xpath('.//domain-name')
        if len(domain_name) > 0:
            facts['domain'] = domain_name[0].text
            facts['fqdn'] = facts['hostname'] + '.' + facts['domain']
            return
    except RpcError:
        pass

    fs = FS(junos)
    file_content = fs.cat('/etc/resolv.conf') or fs.cat('/var/etc/resolv.conf')
    words = file_content.split() if file_content is not None else ''
    if 'domain' not in words:
        facts['domain'] = None
        facts['fqdn'] = facts['hostname']
    else:
        idx = words.index('domain') + 1
        facts['domain'] = words[idx]
        facts['fqdn'] = facts['hostname'] + '.' + facts['domain']



def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {'_iri_hostname': 'A dictionary keyed by internal routing instance '
                             'ip addresses. The value of each key is the '
                             'internal routing instance hostname for the ip',
            '_iri_ip': 'A dictionary keyed by internal routing instance '
                       'hostnames. The value of each key is the internal '
                       'routing instance ip for the hostname', }


def get_facts(device):
    """
    Gathers facts from a <file-show/> RPC on the '/etc/hosts.junos' file.
    """
    iri_hostname = None
    iri_ip = None

    rsp = device.rpc.file_show(filename='/etc/hosts.junos', normalize=False)

    if rsp is not None:
        hosts_file_content = rsp.findtext('.', default='')
        if hosts_file_content is not None:
            for line in hosts_file_content.splitlines():
                (line, _, _) = line.partition('#')
                components = line.split(None)
                if len(components) > 1:
                    ip = components[0]
                    hosts = components[1:]
                    if iri_hostname is None:
                        iri_hostname = {}
                    if iri_ip is None:
                        iri_ip = {}
                    if ip in iri_hostname:
                        iri_hostname[ip] += hosts
                    else:
                        iri_hostname[ip] = hosts
                    for host in hosts:
                        if host in iri_ip:
                            iri_ip[host].append(ip)
                        else:
                            iri_ip[host] = [ip]
                    for host in hosts:
                        # Handle templates with %d
                        if '%d' in host:
                            octets = ip.split('.', 3)
                            for count in range(255):
                                t_ip = (octets[0] + '.' + octets[1] +
                                        '.' + str(count) + '.' + octets[3])
                                t_host = host.replace('%d', str(count))
                                if t_ip in iri_hostname:
                                    iri_hostname[t_ip].append(t_host)
                                else:
                                    iri_hostname[t_ip] = [t_host]
                                if t_host in iri_ip:
                                    iri_ip[t_host].append(t_ip)
                                else:
                                    iri_ip[t_host] = [t_ip]

    return {'_iri_hostname': iri_hostname,
            '_iri_ip': iri_ip, }



from jnpr.junos.exception import RpcError


def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {'_is_linux': "A boolean indicating if the device is running linux"
                      "kernel.", }


def get_facts(device):
    """
    Gathers the _is_linux fact using the <file-show/> RPC on the EVO version file.
    """

    # Temporary implementation until PR 1245634 is implemented.
    LINUX_VERSION_PATH = '/usr/share/cevo/cevo_version'

    is_linux = None

    try:
        rsp = device.rpc.file_show(normalize=True, filename=LINUX_VERSION_PATH)
        if rsp.tag == 'file-content':
            is_linux = True
        else:
            is_linux = False
    except RpcError:
        is_linux = False
    return {'_is_linux': is_linux, }



VERSION = "2.3.0"
DATE = "2019-Sep-27"

# Augment with the internal version if present
try:
    from jnpr.junos.internal_version import INTERNAL_VERSION
    VERSION += '+internal.' + str(INTERNAL_VERSION)
except ImportError:
    # No internal version
    pass



import re
from jnpr.junos.facts.swver import version_info


def _get_swver(dev, facts):
    # See if we're VC Capable
    if facts['vc_capable'] is True:
        try:
            return dev.rpc.cli("show version all-members", format='xml')
        except:
            pass
    try:
        return dev.rpc.cli("show version invoke-on all-routing-engines",
                           format='xml')
    except:
        return dev.rpc.get_software_information()


def facts_software_version(junos, facts):
    """
    The following facts are required:
        facts['master']

    The following facts are assigned:
        facts['hostname']
        facts['version']
        facts['version_<RE#>'] for each RE in dual-RE, cluster or VC system
        facts['version_info'] for master RE
    """

    x_swver = _get_swver(junos, facts)

    if not facts.get('model'):
        # try to extract the model from the version information
        facts['model'] = x_swver.findtext('.//product-model')

    # ------------------------------------------------------------------------
    # extract the version information out of the RPC response
    # ------------------------------------------------------------------------

    f_master = facts.get('master', 'RE0')

    if x_swver.tag == 'multi-routing-engine-results':
        # we need to find/identify each of the routing-engine (CPU) versions.
        if len(x_swver.xpath('./multi-routing-engine-item')) > 1:
            facts['2RE'] = True
        versions = []

        if isinstance(f_master, list):
            xpath = './multi-routing-engine-item[re-name="{}"]/software-' \
                    'information/host-name'.format(f_master[0].lower())
        else:
            xpath = './multi-routing-engine-item[re-name="{}"' \
                    ']/software-information/host-name'.format(f_master.lower())

        facts['hostname'] = x_swver.findtext(xpath)
        if facts['hostname'] is None:
            # then there the re-name is not what we are expecting; we should
            # handle this better, eh?  For now, just assume there is one
            # software-information element and take that host-name. @@@ hack.
            facts['hostname'] = x_swver.findtext(
                './/software-information/host-name')

        for re_sw in x_swver.xpath('.//software-information'):

            re_name = re_sw.xpath('preceding-sibling::re-name')[0].text

            # handle the cases where the "RE name" could be things like
            # "FPC<n>" or "ndoe<n>", and normalize to "RE<n>".
            re_name = re.sub(r'(\w+)(\d+)', 'RE\\2', re_name)

            # First try the <junos-version> tag present in >= 15.1
            swinfo = re_sw.findtext('junos-version', default=None)
            if not swinfo:
                # For < 15.1, get version from the "junos" package.
                pkginfo = re_sw.xpath(
                    'package-information[normalize-space(name)='
                    '"junos"]/comment')[0].text
                try:
                    swinfo = re.findall(r'\[(.*)\]', pkginfo)[0]
                except:
                    swinfo = "0.0I0.0"
            versions.append((re_name.upper(), swinfo))

        # now add the versions to the facts <dict>
        for re_ver in versions:
            facts['version_' + re_ver[0]] = re_ver[1]

        if f_master is not None:
            master = f_master[0] if isinstance(f_master, list) else f_master
            if 'version_' + master in facts:
                facts['version'] = facts['version_' + master]
            else:
                facts['version'] = versions[0][1]
        else:
            facts['version'] = versions[0][1]

    else:
        # single-RE
        facts['hostname'] = x_swver.findtext('host-name')

        # First try the <junos-version> tag present in >= 15.1
        swinfo = x_swver.findtext('.//junos-version', default=None)
        if not swinfo:
            # For < 15.1, get version from the "junos" package.
            pkginfo = x_swver.xpath(
                './/package-information[normalize-space(name)="junos"]/comment'
            )[0].text
            try:
                swinfo = re.findall(r'\[(.*)\]', pkginfo)[0]
            except:
                swinfo = "0.0I0.0"
        facts['version'] = swinfo

    # ------------------------------------------------------------------------
    # create a 'version_info' object based on the master version
    # ------------------------------------------------------------------------

    facts['version_info'] = version_info(facts['version'])



def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {'2RE': "A boolean indicating if the device has more than one "
                   "Routing Engine installed.",
            'master': "On a single chassis/node system, a string value of "
                      "'RE0' or 'RE1' indicating which RE is master. On a "
                      "multi-chassis or multi-node system, the value is a "
                      "list of these strings indicating whether RE0 or RE1 "
                      "is master. There is one entry in the list for each "
                      "chassis/node in the system.",
            'RE0': "A dictionary with information about RE0 (if present). The "
                   "keys of the dictionary are: mastership_state, status, "
                   "model, up_time, and last_reboot_reason.",
            'RE1': "A dictionary with information about RE1 (if present). The "
                   "keys of the dictionary are: mastership_state, status, "
                   "model, up_time, and last_reboot_reason.",
            're_info': "A three-level dictionary with information about "
                       "the Routing Engines in the device. The first-level "
                       "key is the chassis or node name. The second-level key "
                       "is the slot number, the third-level keys are: "
                       "mastership_state, status, model, and "
                       "last_reboot_reason. A first-level key with a value "
                       "of 'default' will always be present and represents "
                       "the first chassis/node of the system (Note: the first "
                       "chasis/node of the system is not necessarily the "
                       "'master' node in a VC.) A second-level key with a "
                       "value of 'default' will always be present "
                       "for the default chassis/node and represents the "
                       "first Routing Engine on the first node/chassis. "
                       "(Note: the first RE of a chassis/node is not "
                       "necessarily the 'master' RE of the chassis/node. See "
                       "the RE_master fact for info on the 'master' RE of "
                       "each chassis/node.)",
            're_master': "A dictionary indicating which RE slot is master for "
                         "each chassis/node in the system. The dictionary key "
                         "is the chassis or node name. A key with a value "
                         "of 'default' will always be present and represents "
                         "the first node/chassis of the system. (Note: the "
                         "first chassis/node of the system is not necessarily "
                         "the 'master' node in a VC. See the vc_master fact "
                         "to determine which chassis/node is the master of "
                         "a VC.)", }


def get_facts(device):
    """
    Gathers facts from the <get-route-engine-information/> RPC.
    """
    multi_re = None
    master = None
    RE0 = None
    RE1 = None
    re_info = None
    re_master = None

    rsp = device.rpc.get_route_engine_information(normalize=True)
    re_list = rsp.findall('.//route-engine')
    if len(re_list) > 1:
        multi_re = True
    else:
        multi_re = False

    first_node = None
    first_slot = None
    master_list = []
    node_masters = {}
    for current_re in re_list:
        node = current_re.findtext('../../re-name', 'default')
        slot = current_re.findtext('slot', '0')
        info = {'mastership_state': current_re.findtext('mastership-state',
                                                        'master'),
                'status': current_re.findtext('status'),
                'model': current_re.findtext('model'),
                'last_reboot_reason':
                    current_re.findtext('last-reboot-reason'),
                # This key is only returned in the RE0 and RE1 facts in order
                # to maintain backwards compatibility with the old fact
                # gathering system. Since the up_time value changes, it's not
                # really a "fact" and is therefore omitted from the new re_info
                # fact.
                'up_time': current_re.findtext('up-time'), }
        if first_node is None:
            first_node = node
            first_slot = slot
        if node == first_node:
            if slot == '0' and RE0 is None:
                # Copy the dictionary
                RE0 = dict(info)
            if slot == '1' and RE1 is None:
                # Copy the dictionary
                RE1 = dict(info)
        # Don't want the up_time key in the new re_info fact.
        if 'up_time' in info:
            del info['up_time']
        if re_info is None:
            re_info = {}
        if node not in re_info:
            re_info[node] = {}
        # Save with second-level key as a string.
        re_info[node][slot] = info
        # If it's a master RE, then save in node_masters and master_list
        if ('mastership_state' in info and
           info['mastership_state'].lower().startswith('master')):
            node_masters[node] = slot
            master_list.append('RE' + slot)
    # Fill in the 'default' first-level key if multi-chassis/node system
    if first_node is not None and first_node != 'default':
        re_info['default'] = re_info[first_node]
        if first_node in node_masters:
            node_masters['default'] = node_masters[first_node]
    # Fill in the 'default' second-level key if at least one RE was found.
    if first_slot is not None:
        re_info['default']['default'] = re_info['default'][first_slot]

    # Set the 'master' fact to a string or list based on the number of members.
    master_list_len = len(master_list)
    if master_list_len == 1:
        master = master_list[0]
    elif master_list_len > 1:
        master = master_list

    # If any info in the node_masters dict, then return it as fact 're_master'
    if node_masters:
        re_master = node_masters

    return {'2RE': multi_re,
            'master': master,
            'RE0': RE0,
            'RE1': RE1,
            're_info': re_info,
            're_master': re_master}



from jnpr.junos.exception import ConnectNotMasterError
from jnpr.junos.exception import RpcError


def facts_chassis(junos, facts):
    """
    The following facts are assigned:
      facts['2RE'] : designates if the device can support two RE, not that it
                     has them
      facts['RE_hw_mi'] : designates if the device is
                          multi-instance-routing-engine
      facts['model'] : product model
      facts['serialnumber'] : serial number

    NOTES:
        (1) if in a 2RE system, this routine will only load the information
            from the first chassis item.
        (2) hostname, domain, and fqdn are retrieved from configuration data;
            inherited configs are checked.
    """
    # Set default values.
    facts['2RE'] = False
    facts['RE_hw_mi'] = False
    facts['model'] = 'UNKNOWN'
    facts['serialnumber'] = 'UNKNOWN'

    rsp = junos.rpc.get_chassis_inventory()
    if rsp.tag == 'error':
        raise RuntimeError()

    if rsp.tag == 'output':
        # this means that there was an error; due to the
        # fact that this connection is not on the master
        # @@@ need to validate on VC-member
        raise ConnectNotMasterError(junos)

    if rsp.tag == 'multi-routing-engine-results':
        facts['2RE'] = True
        facts['RE_hw_mi'] = True
    else:
        facts['2RE'] = False

    facts['model'] = rsp.findtext('.//chassis[1]/description', 'UNKNOWN')
    facts['serialnumber'] = (
        rsp.findtext('.//chassis[1]/serial-number') or
        rsp.findtext('.//chassis-module[name="Backplane"]/serial-number') or
        rsp.findtext('.//chassis-module[name="Midplane"]/serial-number',
                     'UNKNOWN'))

    if facts['model'] == 'UNKNOWN' or facts['serialnumber'] == 'UNKNOWN':
        raise RpcError()



import re


def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {'personality': 'A string which is generally based on the '
                           'platform and indicates the behavior of the '
                           'device.',
            'virtual': 'A boolean indicating if the device is virtual.', }


def get_facts(device):
    """
    Determines personality fact based on the model.
    """
    personality = None
    virtual = None

    model = device.facts['model']

    if model == 'Virtual Chassis':
        # Set model to the model of the first RE in the multi-chassis system.
        model = device.facts['re_info']['default']['default']['model']

    if re.match('^(EX)|(QFX)', model):
        personality = 'SWITCH'
        virtual = False
    elif model.startswith('MX'):
        re_type = device.facts['re_info']['default']['default']['model']
        # The VMX has an RE type of 'RE-VMX'
        if re_type == 'RE-VMX':
            personality = 'MX'
            virtual = True
        # An MX GNF has an RE type that includes the letters 'GNF'
        elif re_type is not None and 'GNF' in re_type:
            personality = 'MX-GNF'
            virtual = True
        else:
            personality = 'MX'
            virtual = False
    elif model.startswith('VMX'):
        personality = 'MX'
        virtual = True
    elif model.startswith('VJX'):
        personality = 'SRX_BRANCH'
        virtual = True
    elif 'VRR' == model:
        personality = 'MX'
        virtual = True
    elif model.startswith('M'):
        personality = 'M'
        virtual = False
    elif model.startswith('T'):
        personality = 'T'
        virtual = False
    elif model.startswith('PTX'):
        personality = 'PTX'
        # The vPTX has an RE type of 'RE-VIRTUAL'
        if (device.facts['re_info']['default']['default']['model'] ==
           'RE-VIRTUAL'):
            virtual = True
        else:
            virtual = False
    elif re.match('SRX\s?(\d){4}', model):
        srx_model = int(model[-4:])
        if srx_model > 5000:
            personality = 'SRX_HIGHEND'
        else:
            personality = 'SRX_MIDRANGE'
        virtual = False
    elif re.match('SRX\s?(\d){3}', model):
        personality = 'SRX_BRANCH'
        virtual = False
    elif re.search('firefly', model, re.IGNORECASE):
        personality = 'SRX_BRANCH'
        virtual = True
    elif 'OLIVE' == model:
        personality = 'OLIVE'
        virtual = True
    elif model.startswith('NFX'):
        personality = 'NFX'
        virtual = False
    elif 'JUNOS_NODE_SLICING' == model:
        personality = 'JDM'
        virtual = True

    return {'personality': personality,
            'virtual': virtual, }



from lxml import etree
from jnpr.junos.exception import PermissionError
from jnpr.junos.utils.fs import FS


def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {'domain': "The domain name configured at the [edit system "
                      "domain-name] configuration hierarchy.",
            'fqdn': "The device's hostname + domain", }


def get_facts(device):
    """
    Gathers domain facts from the configuration or /etc/resolv.conf.
    """
    domain_config = """
        <configuration>
            <system>
                <domain-name/>
             </system>
        </configuration>
    """
    domain = None
    fqdn = None
    # Try to read the domain-name from the config.
    # This might fail due to lack of permissions.
    try:
        rsp = device.rpc.get_config(filter_xml=etree.XML(domain_config),
                                    options={'database': 'committed',
                                             'inherit': 'inherit',
                                             'commit-scripts': 'apply', })
        domain = rsp.findtext('.//domain-name')
    # Ignore if user can't view the configuration.
    except PermissionError:
        pass

    # Try to read the domain from the resolv.conf file. This only requires
    # view permissions.
    if domain is None:
        fs = FS(device)
        file_content = (fs.cat('/etc/resolv.conf') or
                        fs.cat('/var/etc/resolv.conf'))
        words = file_content.split() if file_content is not None else []
        if 'domain' in words:
            idx = words.index('domain') + 1
            domain = words[idx]

    # Set the fqdn
    fqdn = device.facts['hostname']
    if fqdn is not None and domain is not None:
        fqdn = fqdn + '.' + domain

    return {'domain': domain,
            'fqdn': fqdn, }



from select import select
import re
import datetime
from jnpr.junos.utils.ssh_client import open_ssh_client

_JUNOS_PROMPT = '> '
_SHELL_PROMPT = '(%|#|\$)\s'
_SELECT_WAIT = 0.1
_RECVSZ = 1024


class StartShell(object):

    """
    Junos shell execution utility.  This utility is written to
    support the "context manager" design pattern.  For example::

        def _ssh_exec(self, command):
            with StartShell(self._dev) as sh:
                got = sh.run(command)
            return got

    """

    def __init__(self, nc, timeout=30):
        """
        Utility Constructor

        :param Device nc: The Device object

        :param int timeout:
          Timeout value in seconds to wait for expected string/pattern.
        """
        self._nc = nc
        self.timeout = timeout
        self._client = None
        self._chan = None

    def wait_for(self, this=_SHELL_PROMPT, timeout=0):
        """
        Wait for the result of the command, expecting **this** prompt.

        :param str this: expected string/pattern.

        :param int timeout:
          Timeout value in seconds to wait for expected string/pattern.
          If not specified defaults to self.timeout.

        :returns: resulting string of data in a list
        :rtype: list

        .. warning:: need to add a timeout safeguard
        """
        chan = self._chan
        got = []
        timeout = timeout or self.timeout
        timeout = datetime.datetime.now() + datetime.timedelta(
            seconds=timeout)
        while timeout > datetime.datetime.now():
            rd, wr, err = select([chan], [], [], _SELECT_WAIT)
            if rd:
                data = chan.recv(_RECVSZ)
                if isinstance(data, bytes):
                    data = data.decode('utf-8', 'replace')
                got.append(data)
                if this is not None and re.search(r'{}\s?$'.format(this),
                                                  data):
                    break
        return got

    def send(self, data):
        """
        Send the command **data** followed by a newline character.

        :param str data: the data to write out onto the shell.
        :returns: result of SSH channel send
        """
        self._chan.send(data)
        self._chan.send('\n')

    def open(self):
        """
        Open an ssh-client connection and issue the 'start shell' command to
        drop into the Junos shell (csh).  This process opens a
        :class:`paramiko.SSHClient` instance.
        """
        self._client = open_ssh_client(dev=self._nc)
        self._chan = self._client.invoke_shell()

        got = self.wait_for(r'(%|>|#|\$)')
        if got[-1].endswith(_JUNOS_PROMPT):
            self.send('start shell')
            self.wait_for(_SHELL_PROMPT)

    def close(self):
        """ Close the SSH client channel """
        self._chan.close()
        self._client.close()

    def run(self, command, this=_SHELL_PROMPT, timeout=0):
        """
        Run a shell command and wait for the response.  The return is a
        tuple. The first item is True/False if exit-code is 0.  The second
        item is the output of the command.

        :param str command: the shell command to execute
        :param str this: the expected shell-prompt to wait for. If ``this`` is
          set to None, function will wait for all the output on the shell till
          timeout value.
        :param int timeout:
          Timeout value in seconds to wait for expected string/pattern (this).
          If not specified defaults to self.timeout. This timeout is specific
          to individual run call. If ``this`` is provided with None value,
          function will wait till timeout value to grab all the content from
          command output.

        :returns: (last_ok, result of the executed shell command (str) )

        .. code-block:: python

           with StartShell(dev) as ss:
               print ss.run('cprod -A fpc0 -c "show version"', timeout=10)

        .. note:: as a *side-effect* this method will set the ``self.last_ok``
                  property.  This property is set to ``True`` if ``$?`` is
                  "0"; indicating the last shell command was successful else
                  False. If ``this`` is set to None, last_ok will be set to
                  True if there is any content in result of the executed shell
                  command.
        """
        timeout = timeout or self.timeout
        # run the command and capture the output
        self.send(command)
        got = ''.join(self.wait_for(this, timeout))
        self.last_ok = False
        if this is None:
            self.last_ok = got is not ''
        elif this != _SHELL_PROMPT:
            self.last_ok = re.search(r'{}\s?$'.format(this), got) is not None
        elif re.search(r'{}\s?$'.format(_SHELL_PROMPT), got) is not None:
            # use $? to get the exit code of the command
            self.send('echo $?')
            rc = ''.join(self.wait_for(_SHELL_PROMPT))
            self.last_ok = rc.find('\r\n0\r\n') > 0
        return (self.last_ok, got)

    # -------------------------------------------------------------------------
    # CONTEXT MANAGER
    # -------------------------------------------------------------------------

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_ty, exc_val, exc_tb):
        self.close()



def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {'ifd_style': "The type of physical interface (ifd) supported by "
                         "the device. Choices are 'CLASSIC' or 'SWITCH'.", }


def get_facts(device):
    """
    Determines ifd_style fact based on the personality.
    """
    ifd_style = 'CLASSIC'

    if device.facts['personality'] == 'SWITCH':
        ifd_style = 'SWITCH'
    elif device.facts['personality'] == 'JDM':
        ifd_style = None

    return {'ifd_style': ifd_style, }



from jnpr.junos.exception import RpcError


def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {'vc_capable': "A boolean indicating if the device is currently "
                          "configured in a virtual chassis. In spite of the "
                          "name, this fact does NOT indicate whether or not "
                          "the device is CAPABLE of joining a VC.",
            'vc_mode': "A string indicating the current virtual chassis "
                       "mode of the device.",
            'vc_fabric': "A boolean indicating if the device is currently in "
                         "fabric mode.",
            'vc_master': "A string indicating the chassis/node which is "
                         "currently the master of the VC.", }


def get_facts(device):
    """
    Gathers facts from the <get-virtual-chassis-information/> RPC.
    """
    vc_capable = None
    vc_mode = None
    vc_fabric = None
    vc_master = None

    try:
        rsp = device.rpc.get_virtual_chassis_information(normalize=True)
        # MX issue where command returns, but without content. In this case,
        # rsp is set to True.
        if rsp is not True:
            vc_capable = True
            if rsp is not None:
                vc_mode = rsp.findtext('.//virtual-chassis-mode')
                vc_id_info = rsp.find('.//virtual-chassis-id-information')
                if vc_id_info is not None:
                    if vc_id_info.get('style') == 'fabric':
                        vc_fabric = True
                    else:
                        vc_fabric = False
                for member_id in rsp.xpath(
                        ".//member-role[starts-with(.,'Master')]"
                        "/preceding-sibling::member-id"):
                    if vc_master is None:
                        vc_master = member_id.text
                    else:
                        old_vc_master = vc_master
                        vc_master = None
                        raise ValueError("Member %s and member %s both claim "
                                         "to be master of the VC." %
                                         (old_vc_master, member_id.text))
        else:
            vc_capable = False
    except RpcError:
        # Likely a device that doesn't implement the
        # <get-virtual-chassis-information/> RPC.
        # That's OK. Set vc_capable = False.
        vc_capable = False

    return {'vc_capable': vc_capable,
            'vc_mode': vc_mode,
            'vc_fabric': vc_fabric,
            'vc_master': vc_master, }



from jnpr.junos.exception import RpcError


def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {'srx_cluster': "A boolean indicating if the device is part of an "
                           "SRX cluster.",
            'srx_cluster_id': "A string containing the configured cluster id",
            'srx_cluster_redundancy_group': "A multi-level dictionary of "
                                            "information about the SRX "
                                            "cluster redundancy groups "
                                            "on the device. The first-level "
                                            "key is the redundancy group id. "
                                            "The second-level keys are: "
                                            "cluster_id, failover_count, "
                                            "node0, and node1. The node0 and "
                                            "node1 keys have third-level keys "
                                            "of priority, preempt, status, "
                                            "and failover_mode. The values "
                                            "for this fact correspond to the "
                                            "values of the 'show chassis "
                                            "cluster status' CLI command."}


def get_facts(device):
    """
    Gathers facts from the <get-chassis-cluster-status/> RPC.
    """
    srx_cluster = None
    srx_cluster_id = None
    redundancy_group = None

    try:
        rsp = device.rpc.get_chassis_cluster_status(normalize=True)
        if rsp is not None:
            if rsp.tag == 'error':
                srx_cluster = False
            else:
                srx_cluster = True
                srx_cluster_id = rsp.findtext('cluster-id')
                groups = rsp.findall('redundancy-group')
                if groups is not None:
                    redundancy_group = {}
                    for group in groups:
                        group_id = group.findtext('redundancy-group-id')
                        redundancy_group[group_id] = {
                            'cluster_id': group.findtext('cluster-id'),
                            'failover_count': group.findtext(
                                'redundancy-group-failover-count'),
                        }
                        # Iterate over the broken XML in <device-stats>
                        for stats in group.findall('./device-stats'):
                            for node in zip(stats.findall('device-name'),
                                            stats.findall('device-priority'),
                                            stats.findall(
                                                'redundancy-group-status'),
                                            stats.findall('preempt'),
                                            stats.findall('failover-mode')):
                                redundancy_group[group_id][node[0].text] = {
                                    'priority': node[1].text,
                                    'status': node[2].text,
                                    'preempt': node[3].text,
                                    'failover_mode': node[4].text,
                                }
    except RpcError:
        # Likely a device that doesn't implement the
        # <get-chassis-cluster-status/> RPC.
        # That's OK. Just ignore it and leave srx_cluster = None.
        pass
    return {'srx_cluster': srx_cluster,
            'srx_cluster_id': srx_cluster_id,
            'srx_cluster_redundancy_group': redundancy_group}
